/*
	Copyright (C) 1995-2001 AROS - The Amiga Research OS
	$Id: sendintuimessage.c,v 1.6 2010/02/28 14:26:17 jacadcaps Exp $
 
	Desc:
	Lang: english
*/
#include <proto/timer.h>
#include "intuition_intern.h"
#include "icclass.h"
#include "mosmisc.h"

#define	DEBUG_SENDINTUIMESSAGE(x)	;
/*****i***********************************************************************
 
	NAME */
#include <proto/intuition.h>

AROS_LH2(void, SendIntuiMessage,

		 /*  SYNOPSIS */
		 AROS_LHA(struct Window *, window, A0),
		 AROS_LHA(struct IntuiMessage *, imsg, A1),

		 /*  LOCATION */
		 struct IntuitionBase *, IntuitionBase, 151, Intuition)

/*  FUNCTION
	Private: send an IntuiMessage to an Intuition window
 
	INPUTS
	window - The window to which the IntuiMessage shall be sent
		imsg   - The IntuiMessage to send, which must have been allocated with
		 AllocIntuiMessage.
 
	RESULT
		none
 
	NOTES
		The caller of this function should first check himself
	whether window->UserPort is NULL. And in this case do not
	call this function at all.
 
	If inside this function the window->UserPort turns out to
	be NULL, then what happens is, that the IntuiMessage is
	immediately ReplyMessage()ed in here, just like if this was
	done by the app whose window was supposed to get the
	IntuiMessage.
 
	The protection with Forbid() is necessary, because of the
	way shared window userports are handled, when one of this
	windows is closed, where there is also just a protection with
	Forbid() when stripping those IntuiMessages from the port
	which belong to the window which is going to be closed.
 
	This function does not check whether the window to which
	the IntuiMessage is supposed to be sent, really wants to
	get the IDCMP in question, that is, whether the corresponding
	flag in window->IDCMPFLags is set.
 
 
	EXAMPLE
 
	BUGS
 
	SEE ALSO
 
	INTERNALS
 
	HISTORY
 
*****************************************************************************/
{
	AROS_LIBFUNC_INIT
	AROS_LIBBASE_EXT_DECL(struct IntuitionBase *,IntuitionBase)

	ASSERT_VALID_PTR(window);
	ASSERT_VALID_PTR(imsg);

	DEBUG_SENDINTUIMESSAGE(dprintf("SendIntuiMessage: Window 0x%lx Port 0x%lx Msg 0x%lx\n",
	                               window, window->UserPort, imsg));

	SANITY_CHECK(window)
	SANITY_CHECK(imsg)

	Forbid();

#ifdef __MORPHOS__
	CrashCheck(window,IntuitionBase);

	if (IW(window)->specialflags & SPFLAG_IAMDEAD)
	{
		imsg->IDCMPWindow = 0;
		imsg->Code = 0;
		imsg->Qualifier = 0;
		ReplyMsg(&imsg->ExecMessage);

		Permit();

		return;
	}
#endif

	IIM(imsg)->flags = 0;

	if (imsg->Qualifier & IEQUALIFIER_REPEAT)
	{
		IW(window)->num_repeatevents++;
		IIM(imsg)->flags |= INTUIMESSAGEF_REPEATEVENTS;
	}

	switch (imsg->Class)
	{
		case IDCMP_MOUSEMOVE:
			IW(window)->num_mouseevents++;
			IIM(imsg)->flags |= INTUIMESSAGEF_MOUSEEVENTS;
			break;

		case IDCMP_INTUITICKS:
			window->Flags |= WFLG_WINDOWTICKED;
			IIM(imsg)->flags |= INTUIMESSAGEF_INTUITICKS;
			break;

		case IDCMP_MENUVERIFY:
			IS(window->WScreen)->MenuVerifyMsgCount++;
			IIM(imsg)->flags |= INTUIMESSAGEF_MENUVERIFY;
			break;

		case IDCMP_IDCMPUPDATE:
			if (IW(window)->num_idcmpupdate && !IW(window)->messagecache)
			{
				IW(window)->messagecache = imsg;
				Permit();
				return;
			}

			//reduce number of messages if possible (prop updates only!)
			if (IW(window)->num_idcmpupdate && IW(window)->messagecache)
			{
				if (imsg->Code == IW(window)->messagecache->Code && imsg->Qualifier == ICMAGIC)
				{
					IW(window)->messagecache->MouseX = imsg->MouseX;
					IW(window)->messagecache->MouseY = imsg->MouseY;
					IW(window)->messagecache->Seconds = imsg->Seconds;
					IW(window)->messagecache->Micros = imsg->Micros;
					imsg->IDCMPWindow = 0;
					imsg->Code = 0;
					imsg->Qualifier = 0;
					ReplyMsg(&imsg->ExecMessage);
					Permit();
					return;
				}
			}

			imsg->Qualifier = 0;
			IW(window)->num_idcmpupdate++;
			IIM(imsg)->flags |= INTUIMESSAGEF_IDCMPUPDATE;
			break;

		case IDCMP_NEWSIZE:
		case IDCMP_CHANGEWINDOW:
			IW(window)->num_sizeevents ++;
			IIM(imsg)->flags |= INTUIMESSAGEF_SIZEEVENTS;
			break;

		case IDCMP_SIZEVERIFY:
			IIM(imsg)->flags |= INTUIMESSAGEF_SIZEVERIFY;
			break;

		case IDCMP_REFRESHWINDOW:
			IW(window)->num_refreshevents ++;
			IIM(imsg)->flags |= INTUIMESSAGEF_REFRESHVERIFY;
			break;

		case IDCMP_REQVERIFY:
			IIM(imsg)->flags |= INTUIMESSAGEF_REQVERIFY;
			break;
	}

	IW(window)->lastmsgsent.tv_secs = IntuitionBase->Seconds;
	IW(window)->lastmsgsent.tv_micro = IntuitionBase->Micros;

	if (window->UserPort)
	{
		DEBUG_SENDINTUIMESSAGE(dprintf("SendIntuiMessage: Class 0x%lx Code 0x%lx Qual 0x%lx Mouse %d,%d IAddress 0x%lx\n", imsg->Class, imsg->Code, imsg->Qualifier, imsg->MouseX, imsg->MouseY, imsg->IAddress));

		IIM(imsg)->origwindow = imsg->IDCMPWindow;

		PutMsg(window->UserPort, &imsg->ExecMessage);

		/* Help sucky programs... (die DigiBooster, die!) */
		window->MessageKey = imsg;
	} else {
		ReplyMsg(&imsg->ExecMessage);
	}

	Permit();

	AROS_LIBFUNC_EXIT
}

